@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1999, 2000, 2001, 2002, 2003, 2004 SuSE Linux AG, Germany.
+ * Copyright (c) 2006 SuSE Linux Products GmbH, Germany.
+ * Copyright (c) 1999-2004 SuSE Linux AG, Germany.
  * Author: Thorsten Kukuk <kukuk@suse.de>
  *
  * Redistribution and use in source and binary forms, with or without
@@ -38,8 +39,6 @@
 #include "config.h"
 #endif
 
-#define _GNU_SOURCE
-
 #include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -53,12 +52,15 @@
 #include <sys/time.h>
 #include <rpc/types.h>
 #include <nss.h>
+#include <syslog.h>
 #include <rpcsvc/yp_prot.h>
 #include <rpcsvc/ypclnt.h>
 
 #define PAM_SM_PASSWORD
-#include <security/_pam_macros.h>
 #include <security/pam_modules.h>
+#if defined (HAVE_SECURITY_PAM_EXT_H)
+#include <security/pam_ext.h>
+#endif
 
 #if defined(HAVE_XCRYPT_H)
 #include <xcrypt.h>
@@ -78,9 +80,9 @@
 #endif
 
 #define MAX_LOCK_RETRIES 3 /* How often should we try to lock password file */
-#define OLD_PASSWORD_PROMPT "Old Password: "
-#define NEW_PASSWORD_PROMPT "New password: "
-#define AGAIN_PASSWORD_PROMPT "Re-enter new password: "
+#define OLD_PASSWORD_PROMPT _("Old Password: ")
+#define NEW_PASSWORD_PROMPT _("New Password: ")
+#define AGAIN_PASSWORD_PROMPT _("Reenter New Password: ")
 
 #define SCALE (24L*3600L)
 
@@ -103,21 +105,21 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
   memset (&output, 0, sizeof (output));
   memset (&options, 0, sizeof (options));
 
-  if (get_options (&options, "password", argc, argv) < 0)
+  if (get_options (pamh, &options, "password", argc, argv) < 0)
     {
-      __pam_log (LOG_ERR, "cannot get options");
-      return PAM_BUF_ERR;
+      pam_syslog (pamh, LOG_ERR, "cannot get options");
+      return PAM_SYSTEM_ERR;
     }
 
-
   if (options.debug)
-    __pam_log (LOG_DEBUG, "pam_sm_chauthtok() called");
+    pam_syslog (pamh, LOG_DEBUG, "pam_sm_chauthtok() called");
 
   /* get the user name */
   if ((retval = pam_get_user (pamh, &name, NULL)) != PAM_SUCCESS)
     {
       if (options.debug)
-	__pam_log (LOG_DEBUG, "pam_get_user failed: return %d", retval);
+	pam_syslog (pamh, LOG_DEBUG,
+		    "pam_get_user failed: return %d", retval);
       return (retval == PAM_CONV_AGAIN ? PAM_INCOMPLETE:retval);
     }
 
@@ -125,23 +127,15 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
     {
       if (name)
 	{
-	  __pam_log (LOG_ERR, "bad username [%s]", name);
+	  pam_syslog (pamh, LOG_ERR, "bad username [%s]", name);
 	  return PAM_USER_UNKNOWN;
 	}
       else if (options.debug)
-	__pam_log (LOG_DEBUG, "name == NULL, return PAM_SERVICE_ERR");
+	pam_syslog (pamh, LOG_DEBUG, "name == NULL, return PAM_SERVICE_ERR");
       return PAM_SERVICE_ERR;
     }
   else if (options.debug)
-    __pam_log (LOG_DEBUG, "username=[%s]", name);
-
-  data = __do_getpwnam (name, options.nisdir);
-  if (data == NULL || data->service == S_NONE)
-    {
-      if (data)
-	free_user_t (data);
-      return PAM_USER_UNKNOWN;
-    }
+    pam_syslog (pamh, LOG_DEBUG, "username=[%s]", name);
 
   /* If we use ldap, handle pam_ldap like "sufficient". If it returns
      success, we should also return success. Else ignore the call.  */
@@ -158,64 +152,28 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
 				"pam_sm_chauthtok",
 				&options);
 	  if (retval == PAM_SUCCESS)
-	    {
-	      if (data)
-		free_user_t (data);
-	      return retval;
-	    }
+	    return retval;
 	}
     }
 
-  if (flags & PAM_PRELIM_CHECK)
+  data = __do_getpwnam (name, options.nisdir);
+  if (data == NULL || data->service == S_NONE)
     {
-      /* This indicates that the modules are being probed as to their
-         ready status for altering the user's authentication token. If
-         the module requires access to another system over some network
-         it should attempt to verify it can connect to this system on
-         receiving this flag. If a module cannot establish it is ready to
-         update the user's authentication token it should return
-         PAM_TRY_AGAIN, this information will be passed back to the
-         application. */
-      retval = PAM_SUCCESS;
-
-      if (data->is_expired)
-	retval = PAM_ACCT_EXPIRED;
-
       if (data)
 	free_user_t (data);
-      return retval;
+      return PAM_USER_UNKNOWN;
     }
 
-  /* Now we have all the initial information we need from the app to
-     set things up (we assume that getting the username succeeded...) */
-  if ((flags & PAM_CHANGE_EXPIRED_AUTHTOK) || getuid () != 0)
-    {
-      if (data->is_tooearly)
-	{
-	  if (data->use_hp_aging)
-	    __write_message (pamh, flags, PAM_TEXT_INFO,
-			     _("Sorry: less then %d weeks since the last change."),
-			     data->hp_week);
-	  else
-	    __write_message (pamh, flags, PAM_TEXT_INFO,
-			     _("Sorry: less then %d days since the last change."),
-			     data->spw.sp_min);
-	  free_user_t (data);
-	  return PAM_AUTHTOK_ERR;
-	}
-      else if (data->is_expired)
-	{
-	  free_user_t (data);
-	  return PAM_ACCT_EXPIRED; /* If their account has expired, we
-				      can't auth them to change their
-				      password */
-	}
-    }
-  if ((data->use_shadow || data->use_hp_aging) && !data->is_expiring &&
-      (flags & PAM_CHANGE_EXPIRED_AUTHTOK))
+  /* Before we do anything else, check to make sure that the user's
+     info is in one of the databases we can modify from this module,
+     which currently is 'files' and 'nis'. */
+  if (data->service != S_LOCAL && data->service != S_YP)
     {
       free_user_t (data);
-      return PAM_SUCCESS;
+      pam_syslog (pamh, LOG_DEBUG,
+		  "user \"%s\" does not exist in /etc/passwd or NIS",
+		  name);
+      return PAM_USER_UNKNOWN;
     }
 
   retval = pam_get_item (pamh, PAM_OLDAUTHTOK, (const void **) &oldpass_void);
@@ -232,9 +190,8 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
 	 we don't need to get the old password. The test for
 	 PAM_CHANGE_EXPIRED_AUTHTOK is here, because login runs as
          root and we need the old password in this case.  */
-
       if (options.debug)
-	__pam_log (LOG_DEBUG, "No old password found.");
+	pam_syslog (pamh, LOG_DEBUG, "No old password found.");
 
       /* Ask for old password: if we are a normal user or if
 	 we are called by login or if we are not a local user.
@@ -249,13 +206,13 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
 	  if (options.use_first_pass)
 	    {
 	      if (options.debug)
-		__pam_log (LOG_DEBUG, "use_first_pass set -> abort.");
+		pam_syslog (pamh, LOG_DEBUG, "use_first_pass set -> abort.");
 
 	      free_user_t (data);
 	      if (getuid () == 0 && data->service != S_LOCAL)
 		__write_message (pamh, flags, PAM_ERROR_MSG,
 				 _("You can only change local passwords."));
-	      return PAM_AUTHTOK_RECOVER_ERR;
+	      return PAM_AUTHTOK_RECOVERY_ERR;
 	    }
 
 	  retval = __get_passwd (pamh, OLD_PASSWORD_PROMPT, &tmp_oldpass);
@@ -268,7 +225,8 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
 	  if (tmp_oldpass == NULL)
 	    {
 	      if (options.debug)
-		__pam_log (LOG_DEBUG, "old password is NULL -> abort.");
+		pam_syslog (pamh, LOG_DEBUG,
+			    "old password is NULL -> abort.");
 	      free_user_t (data);
 	      return PAM_AUTH_ERR;
 	    }
@@ -281,28 +239,73 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
 	      && !options.nullok)
 	    {
 	      if (options.debug)
-		__pam_log (LOG_DEBUG, "old password is empty which is not allowed -> abort.");
+		pam_syslog (pamh, LOG_DEBUG,
+		      "old password is empty which is not allowed -> abort.");
 	      free_user_t (data);
 	      return PAM_AUTH_ERR;
 	    }
 
-	  if (strcmp (data->oldpassword,
-		      crypt_r (oldpass, data->oldpassword, &output)) != 0)
-	    {
-	      if (options.debug)
-		__pam_log (LOG_DEBUG, "old password is wrong -> abort.");
-
-	      free_user_t (data);
-	      return PAM_AUTH_ERR;
-	    }
-
-	  if (!options.not_set_pass)
-	    pam_set_item (pamh, PAM_OLDAUTHTOK, (void *) oldpass);
+	  pam_set_item (pamh, PAM_OLDAUTHTOK, (void *) oldpass);
 	}
       else
 	oldpass = "";
     }
 
+  if (flags & PAM_PRELIM_CHECK)
+    {
+      /* Check if the old password was correct.  */
+      if (getuid () && strcmp (data->oldpassword,
+			crypt_r (oldpass, data->oldpassword, &output)) != 0)
+	{
+	  if (options.debug)
+	    pam_syslog (pamh, LOG_DEBUG,
+			"old password is wrong -> abort.");
+
+	  retval = PAM_AUTH_ERR;
+	}
+      else if (data->is_expired)
+	retval = PAM_ACCT_EXPIRED;
+      else
+	retval = PAM_SUCCESS;
+
+      free_user_t (data);
+      return retval;
+    }
+
+
+  /* Now we have all the initial information we need from the app to
+     set things up (we assume that getting the username succeeded...) */
+  if ((flags & PAM_CHANGE_EXPIRED_AUTHTOK) || getuid () != 0)
+    {
+      if (data->is_tooearly)
+	{
+	  if (data->use_hp_aging)
+	    __write_message (pamh, flags, PAM_TEXT_INFO,
+			     _("Less then %d weeks since the last change."),
+			     data->hp_week);
+	  else
+	    __write_message (pamh, flags, PAM_TEXT_INFO,
+			     _("Less then %d days since the last change."),
+			     data->spw.sp_min);
+	  free_user_t (data);
+	  return PAM_AUTHTOK_ERR;
+	}
+      else if (data->is_expired)
+	{
+	  free_user_t (data);
+	  return PAM_ACCT_EXPIRED; /* If their account has expired, we
+				      can't auth them to change their
+				      password */
+	}
+    }
+
+  if ((data->use_shadow || data->use_hp_aging) && !data->is_expiring &&
+      (flags & PAM_CHANGE_EXPIRED_AUTHTOK))
+    {
+      free_user_t (data);
+      return PAM_SUCCESS;
+    }
+
   /* create a copy which we can give free later */
   data->oldclearpwd = strdup (oldpass);
 
@@ -363,7 +366,7 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
       else
 	{
 	  __write_message (pamh, flags, PAM_ERROR_MSG,
-			   _("Sorry, passwords do not match."));
+			   _("Passwords do not match."));
 	  if (new2 != NULL)
 	    {
 	      memset (new2, '\0', strlen (new2));
@@ -393,8 +396,7 @@ pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc, const char **argv)
     }
   else
     {
-      if (!options.not_set_pass)
-	pam_set_item (pamh, PAM_AUTHTOK, (void *) newpass);
+      pam_set_item (pamh, PAM_AUTHTOK, (void *) newpass);
 
       free_user_t (data);
       __write_message (pamh, flags, PAM_TEXT_INFO, _("Password changed."));
@@ -414,7 +416,7 @@ getnismaster (pam_handle_t *pamh, int flags)
   if ((err = yp_master (domainname, "passwd.byname", &master)) != 0)
     {
       __write_message (pamh, flags, PAM_ERROR_MSG,
-		       _("can't find the master ypserver: %s"),
+		       _("Cannot find the master ypserver: %s"),
 		       yperr_string (err));
       return NULL;
     }
@@ -479,7 +481,7 @@ make_crypt_salt (const char *crypt_prefix, int crypt_rounds,
   if (fd < 0)
     {
       __write_message (pamh, flags, PAM_ERROR_MSG,
-		       _("Can't open %s for reading: %s"),
+		       _("Cannot open %s for reading: %s"),
 		       RANDOM_DEVICE, strerror (errno));
       return NULL;
     }
@@ -503,8 +505,8 @@ make_crypt_salt (const char *crypt_prefix, int crypt_rounds,
   if (!retval)
     {
       __write_message (pamh, flags, PAM_ERROR_MSG,
-		       _("Unable to generate a salt, "
-			 "check your crypt settings."));
+		       _("Unable to generate a salt. "
+			 "Check your crypt settings."));
       return NULL;
     }
 
@@ -553,15 +555,6 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	      crypt_str = getlogindefs_str (opt_str, NULL);
 	    }
 	  break;
-	case S_NISPLUS:
-	  opt_str = "CRYPT_NIS";
-	  crypt_str = getlogindefs_str (opt_str, NULL);
-	  if (crypt_str == NULL)
-	    {
-	      opt_str = "CRYPT_NIS+";
-	      crypt_str = getlogindefs_str (opt_str, NULL);
-	    }
-	  break;
 	case S_LDAP:
 	  opt_str = "CRYPT_LDAP";
 	  crypt_str = getlogindefs_str (opt_str, NULL);
@@ -585,7 +578,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 #if defined(HAVE_BIGCRYPT)
 	  options->use_crypt = BIGCRYPT;
 #else
-	  __pam_log (LOG_ERR, "No bigcrypt support compiled in");
+	  pam_syslog (pamh, LOG_ERR, "No bigcrypt support compiled in");
 #endif
 	}
       else if (strcasecmp (crypt_str, "bf") == 0 ||
@@ -604,15 +597,16 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 		options->crypt_rounds = atoi (round); /* XXX strtol */
 	    }
 	  if (options->debug)
-	    __pam_log (LOG_DEBUG, "Blowfish crypt_rounds=%d",
-		       options->crypt_rounds);
+	    pam_syslog (pamh, LOG_DEBUG, "Blowfish crypt_rounds=%d",
+			options->crypt_rounds);
 #else
-	  __pam_log (LOG_ERR, "No blowfish support compiled in");
+	  pam_syslog (pamh, LOG_ERR, "No blowfish support compiled in");
 #endif
 	}
       else
 	{
-	  __pam_log (LOG_ERR, "Don't know \"%s\", fall back to DES", crypt_str);
+	  pam_syslog (pamh, LOG_ERR,
+		      "Don't know \"%s\", fall back to DES", crypt_str);
 	  options->use_crypt = DES;
 	}
       free_getlogindefs_data ();
@@ -668,7 +662,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
       free (salt);
 #else
       __write_message (pamh, flags, PAM_ERROR_MSG,
-		       _("No support for bigcrypt compiled in"));
+		       _("No support for bigcrypt included"));
       return PAM_AUTHTOK_ERR;
 #endif
       break;
@@ -686,7 +680,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
       free (salt);
 #else
       __write_message (pamh, flags, PAM_ERROR_MSG,
-		       _("No support for blowfish compiled in"));
+		       _("No support for blowfish included"));
       return PAM_AUTHTOK_ERR;
 #endif
       break;
@@ -705,7 +699,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
   if (data->service == S_LOCAL || (data->service == S_YP && options->nisdir))
     {
       int retries = 0;
-      char *etcdir;
+      const char *etcdir;
 
       if (options->nisdir && data->service == S_YP)
 	etcdir = options->nisdir;
@@ -748,14 +742,14 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  if ((oldpf = fopen (shadow_orig, "r")) == NULL)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't open %s: %m"), shadow_orig);
+			       _("Cannot open %s: %m"), shadow_orig);
 	      retval = PAM_AUTHTOK_ERR;
 	      goto error_shadow;
 	    }
 	  if (fstat (fileno (oldpf), &passwd_stat) < 0)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't stat %s: %m"), shadow_orig);
+			       _("Cannot stat %s: %m"), shadow_orig);
 	      fclose (oldpf);
 	      retval = PAM_AUTHTOK_ERR;
 	      goto error_shadow;
@@ -763,7 +757,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 
 #ifdef WITH_SELINUX
           security_context_t prev_context;
-          if (set_default_context (shadow_orig, &prev_context) < 0)
+          if (set_default_context (pamh, shadow_orig, &prev_context) < 0)
             {
               fclose (oldpf);
               retval = PAM_AUTHTOK_ERR;
@@ -773,7 +767,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  /* Open a temp shadow file */
 	  newpf_fd = mkstemp (shadow_tmp);
 #ifdef WITH_SELINUX
-          if (restore_default_context (prev_context) < 0)
+          if (restore_default_context (pamh, prev_context) < 0)
 	    {
 	      fclose (oldpf);
 	      retval = PAM_AUTHTOK_ERR;
@@ -783,15 +777,22 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  if (newpf_fd == -1)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't create temp file (%s): %m"),
+			       _("Cannot create temp file (%s): %m"),
 			       shadow_tmp);
 	      fclose (oldpf);
 	      retval = PAM_AUTHTOK_ERR;
 	      goto error_shadow;
 	    }
-	  fchmod (newpf_fd, passwd_stat.st_mode);
-	  fchown (newpf_fd, passwd_stat.st_uid, passwd_stat.st_gid);
-          if (copy_xattr (shadow_orig, shadow_tmp) > 0)
+	  if (fchmod (newpf_fd, passwd_stat.st_mode) == -1)
+	    __write_message (pamh, flags, PAM_ERROR_MSG,
+			     _("Cannot change permssions of %s: %m"),
+			     shadow_tmp);
+	  if (fchown (newpf_fd, passwd_stat.st_uid,
+		      passwd_stat.st_gid) == -1)
+	    __write_message (pamh, flags, PAM_ERROR_MSG,
+			     _("Cannot change owner/group of %s: %m"),
+			     shadow_tmp);
+          if (copy_xattr (pamh, shadow_orig, shadow_tmp) > 0)
             {
               fclose (oldpf);
               close (newpf_fd);
@@ -802,7 +803,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  if (newpf == NULL)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't open %s: %m"),
+			       _("Cannot open %s: %m"),
 			       shadow_tmp);
 	      fclose (oldpf);
 	      close (newpf_fd);
@@ -855,7 +856,10 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
               goto error_passwd;
             }
 	  unlink (shadow_old);
-	  link (shadow_orig, shadow_old);
+	  if (link (shadow_orig, shadow_old) == -1)
+	    __write_message (pamh, flags,PAM_ERROR_MSG,
+			     _("Cannot create backup file of %s: %m"),
+			     shadow_orig);
 	  rename (shadow_tmp, shadow_orig);
 	error_shadow:
 	  unlink (shadow_tmp);
@@ -864,12 +868,12 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	{
 	  const char *fn_passwd_tmp = "/passwd.tmpXXXXXX";
 	  const char *fn_passwd_orig = "/passwd";
-	  char *passwd_tmp =
-	    alloca (strlen (etcdir) + strlen (fn_passwd_tmp) + 2);
 	  char *passwd_orig =
 	    alloca (strlen (etcdir) + strlen (fn_passwd_orig) + 2);
 	  char *passwd_old =
 	    alloca (strlen (etcdir) + strlen (fn_passwd_orig) + 6);
+	  char *passwd_tmp =
+	    alloca (strlen (etcdir) + strlen (fn_passwd_tmp) + 2);
 	  struct stat passwd_stat;
 	  struct passwd *pw; /* passwd struct obtained from fgetpwent() */
 	  FILE *oldpf, *newpf;
@@ -882,14 +886,14 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  if ((oldpf = fopen (passwd_orig, "r")) == NULL)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't open %s: %m"), passwd_orig);
+			       _("Cannot open %s: %m"), passwd_orig);
 	      retval = PAM_AUTHTOK_ERR;
 	      goto error_passwd;
 	    }
           if (fstat (fileno (oldpf), &passwd_stat) < 0)
             {
               __write_message (pamh, flags, PAM_ERROR_MSG,
-                               _("Can't stat %s: %m"), passwd_orig);
+                               _("Cannot stat %s: %m"), passwd_orig);
               fclose (oldpf);
 	      retval = PAM_AUTHTOK_ERR;
               goto error_passwd;
@@ -897,7 +901,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 
 #ifdef WITH_SELINUX
           security_context_t prev_context;
-          if (set_default_context (passwd_orig, &prev_context) < 0)
+          if (set_default_context (pamh, passwd_orig, &prev_context) < 0)
             {
               fclose (oldpf);
               retval = PAM_AUTHTOK_ERR;
@@ -907,7 +911,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  /* Open a temp passwd file */
 	  newpf_fd = mkstemp (passwd_tmp);
 #ifdef WITH_SELINUX
-          if (restore_default_context (prev_context) < 0)
+          if (restore_default_context (pamh, prev_context) < 0)
 	    {
 	      fclose (oldpf);
 	      retval = PAM_AUTHTOK_ERR;
@@ -917,15 +921,21 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  if (newpf_fd == -1)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't create temp file (%s): %m"),
+			       _("Cannot create temp file (%s): %m"),
 			       passwd_tmp);
 	      fclose (oldpf);
 	      retval = PAM_AUTHTOK_ERR;
 	      goto error_passwd;
 	    }
-          fchmod (newpf_fd, passwd_stat.st_mode);
-          fchown (newpf_fd, passwd_stat.st_uid, passwd_stat.st_gid);
-          if (copy_xattr (passwd_orig, passwd_tmp) > 0)
+          if (fchmod (newpf_fd, passwd_stat.st_mode) == -1)
+	    __write_message (pamh, flags, PAM_ERROR_MSG,
+                             _("Cannot change permssions of %s: %m"),
+			     passwd_tmp);
+          if (fchown (newpf_fd, passwd_stat.st_uid, passwd_stat.st_gid) == -1)
+	    __write_message (pamh, flags, PAM_ERROR_MSG,
+                             _("Cannot change owner/group of %s: %m"),
+                             passwd_tmp);
+          if (copy_xattr (pamh, passwd_orig, passwd_tmp) > 0)
             {
               fclose (oldpf);
               close (newpf_fd);
@@ -937,7 +947,7 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  if (newpf == NULL)
 	    {
 	      __write_message (pamh, flags, PAM_ERROR_MSG,
-			       _("Can't open %s: %m"), passwd_tmp);
+			       _("Cannot open %s: %m"), passwd_tmp);
 	      fclose (oldpf);
 	      close (newpf_fd);
 	      retval = PAM_AUTHTOK_ERR;
@@ -986,7 +996,10 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
               goto error_passwd;
             }
 	  unlink (passwd_old);
-	  link (passwd_orig, passwd_old);
+	  if (link (passwd_orig, passwd_old) == -1)
+	    __write_message (pamh, flags,PAM_ERROR_MSG,
+			     _("Cannot create backup file of %s: %m"),
+			     passwd_orig);
 	  rename (passwd_tmp, passwd_orig);
 	error_passwd:
 	  unlink (passwd_tmp);
@@ -1036,16 +1049,6 @@ __do_setpass (pam_handle_t *pamh, int flags, user_t *data,
 	  retval = PAM_AUTHTOK_ERR;
 	}
     }
-  else if (data->service == S_NISPLUS)
-    {
-      retval = npd_upd_pwd (pamh, flags, data);
-      if (retval != 0)
-	{
-	  __write_message (pamh, flags, PAM_ERROR_MSG,
-			   _("Error while changing the NIS+ password."));
-	  retval = PAM_AUTHTOK_ERR;
-	}
-    }
 
   return retval;
 }
